/*
 * delm_cmdu_ackq.c
 * CMDU response and ack queue management
 *
 * Copyright (C) 2020 IOPSYS Software Solutions AB. All rights reserved.
 *
 * Author: anjan.chanda@iopsys.eu
 *
 * See LICENSE file for license related information.
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>

#include <easy/easy.h>

#include "timer.h"
#include "cmdu_ackq.h"

#define err(...)	log_stderr(0, __VA_ARGS__)
#define dbg(...)	log_stderr(3, __VA_ARGS__)
#define loud(...)	log_stderr(5, __VA_ARGS__)

static int timeradd_msecs(struct timeval *a, unsigned long msecs,
			  struct timeval *res)
{
	if (res) {
		struct timeval t = { 0 };

		if (msecs > 1000) {
			t.tv_sec += msecs / 1000;
			t.tv_usec = (msecs % 1000) * 1000;
		} else {
			t.tv_usec = msecs * 1000;
		}

		timeradd(a, &t, res);
		return 0;
	}

	return -1;
}

struct cmdu_ackq_entry *cmdu_ackq_create_msg(uint16_t type, uint16_t mid,
					     uint8_t *dest, uint32_t timeout,
					     int resend_cnt, void *cookie)
{
	struct cmdu_ackq_entry *msg;
	struct timeval tsp = { 0 };

	msg = calloc(1, sizeof(*msg));
	if (!msg) {
		err("calloc failed! -ENOMEM\n");
		return NULL;
	}

	msg->type = type;
	msg->mid = mid;
	gettimeofday(&tsp, NULL);
	msg->ageing_time = timeout;
	timeradd_msecs(&tsp, msg->ageing_time, &msg->ageing_tmo);

	//msg->ageing_tmo.tv_usec = roundup(msg->ageing_tmo.tv_usec, 1000);
	msg->ageing_tmo.tv_usec = (msg->ageing_tmo.tv_usec / 1000) * 1000;
	memcpy(msg->origin, dest, 6);
	msg->resend_cnt = resend_cnt;
	msg->cookie = cookie;
	dbg("    CREATE msg: type = 0x%04x  mid = %hu origin = " MACFMT " timeout = { %u (%lu:%lu) }\n",
		type, mid, MAC2STR(dest), msg->ageing_time,
		msg->ageing_tmo.tv_sec, msg->ageing_tmo.tv_usec / 1000);

	return msg;
}

static void cmdu_ackq_delete_msg(struct cmdu_ackq *st, struct cmdu_ackq_entry *msg)
{
	if (!msg)
		return;

	if (msg->cookie) {
		if (st->delete_cb)
			st->delete_cb(st, msg);
		else
			free(msg->cookie);
	}

	free(msg);
}

static void cmdu_ackq_ageout_entry(struct cmdu_ackq *st, struct hlist_head *head,
				   struct timeval *min_next_tmo)
{
	struct cmdu_ackq_entry *msg;
	struct hlist_node *tmp;
	struct timeval now = { 0 };


	gettimeofday(&now, NULL);
	now.tv_usec = (now.tv_usec / 1000) * 1000;


	hlist_for_each_entry_safe(msg, tmp, head, hlist) {
		int action = CMDU_ACKQ_TMO_NONE;
		struct timeval new_next_tmo = { 0 };

		dbg("%s(): Entry msg->ageout = (%lu.%lu), now = (%lu.%lu)\n",
		     __func__, msg->ageing_tmo.tv_sec, msg->ageing_tmo.tv_usec,
		     now.tv_sec, now.tv_usec);

		if (timercmp(&msg->ageing_tmo, &now, <=)) {
			dbg("No response from " MACFMT " with CMDU 0x%x mid = %hu\n",
				MAC2STR(msg->origin), msg->type, msg->mid);
			if (st->timeout_cb) {
				action = st->timeout_cb(st, msg);
				if (action == CMDU_ACKQ_TMO_REARM) {
					timeradd_msecs(&now, msg->ageing_time,
							&msg->ageing_tmo);
				}
			} else {
				action = CMDU_ACKQ_TMO_DELETE;
			}
		} else {
			action = CMDU_ACKQ_TMO_REARM;
		}

		switch (action) {
		case CMDU_ACKQ_TMO_DELETE:
			st->pending_cnt--;
			hlist_del(&msg->hlist, head);
			cmdu_ackq_delete_msg(st, msg);
			break;
		case CMDU_ACKQ_TMO_REARM:
			timersub(&msg->ageing_tmo, &now, &new_next_tmo);
			if (!timercmp(min_next_tmo, &new_next_tmo, <)) {
				min_next_tmo->tv_sec = new_next_tmo.tv_sec;
				min_next_tmo->tv_usec = new_next_tmo.tv_usec;
				loud("next-tmo = (%lu.%lu)\n",
					min_next_tmo->tv_sec,
					min_next_tmo->tv_usec);
			}
			break;
		}
	}
}

static void cmdu_ackq_ageing_timer_run(atimer_t *t)
{
	struct cmdu_ackq *st = container_of(t, struct cmdu_ackq, ageing_timer);
	//struct timeval *next_tmo = &st->next_tmo;
	struct timeval min_next_tmo = { .tv_sec = 999999 };
	int remain_cnt = 0;
	int i;
	struct timeval nu;

	gettimeofday(&nu, NULL);

	loud("\n ----Timer now = %lu.%lu  --- cnt = %d---\n",
		nu.tv_sec, nu.tv_usec, st->pending_cnt);

	//spin_lock(&st->hash_lock);
	for (i = 0; i < CMDU_BACKLOG_MAX; i++) {
		if (hlist_empty(&st->table[i]))
			continue;

		loud("i = %d\t", i);
		cmdu_ackq_ageout_entry(st, &st->table[i], &min_next_tmo);
	}

	remain_cnt = st->pending_cnt;
	st->next_tmo.tv_sec = min_next_tmo.tv_sec;
	st->next_tmo.tv_usec = min_next_tmo.tv_usec;

	//spin_unlock(&st->hash_lock);
	if (remain_cnt) {
		uint32_t tmo_msecs =
			st->next_tmo.tv_sec * 1000 + st->next_tmo.tv_usec / 1000;

		 loud("%s(): remain = %d next_tmo = {(%lu s, %lu us) %u}\n",
		       __func__, remain_cnt, st->next_tmo.tv_sec, st->next_tmo.tv_usec, tmo_msecs); 

		if (tmo_msecs > 0)
			timer_set(&st->ageing_timer, tmo_msecs);
	}
}

int cmdu_ackq_init(void *cmdu_q)
{
	struct cmdu_ackq *q = (struct cmdu_ackq *)cmdu_q;

	memset(q, 0, sizeof(*q));
	pthread_mutex_init(&q->qlock, NULL);
	timer_init(&q->ageing_timer, cmdu_ackq_ageing_timer_run);

	return 0;
}

struct cmdu_ackq_entry *cmdu_ackq_lookup(void *cmdu_q, uint16_t type,
					 uint16_t mid, uint8_t *dest)
{
	struct cmdu_ackq *q = (struct cmdu_ackq *)cmdu_q;
	int idx = cmdu_ackq_hash(type, mid, dest);
	struct cmdu_ackq_entry *msg = NULL;

	pthread_mutex_lock(&q->qlock);
	hlist_for_each_entry(msg, &q->table[idx], hlist) {
		if (msg->type == type && msg->mid == mid &&
		    !memcmp(msg->origin, dest, 6)) {

			pthread_mutex_unlock(&q->qlock);
			return msg;
		}
	}
	pthread_mutex_unlock(&q->qlock);

	return NULL;
}

void cmdu_ackq_flush(void *cmdu_q)
{
	struct cmdu_ackq *q = (struct cmdu_ackq *)cmdu_q;
	struct cmdu_ackq_entry *msg = NULL;
	int idx = 0;

	pthread_mutex_lock(&q->qlock);
	for (idx = 0; idx < CMDU_BACKLOG_MAX; idx++) {
		hlist_for_each_entry(msg, &q->table[idx], hlist)
			cmdu_ackq_delete_msg(q, msg);

		q->table[idx].first = NULL;
	}

	q->pending_cnt = 0;
	pthread_mutex_unlock(&q->qlock);
}

void cmdu_ackq_free(void *cmdu_q)
{
	struct cmdu_ackq *q = (struct cmdu_ackq *)cmdu_q;

	cmdu_ackq_flush(q);
	timer_del(&q->ageing_timer);
	pthread_mutex_destroy(&q->qlock);
}

/* In this function, type = cmdutype that is expected with 'mid' from 'dest' */
int cmdu_ackq_enqueue(void *cmdu_q, uint16_t type, uint16_t mid, uint8_t *dest,
		      uint32_t timeout, int resend_cnt, void *cookie)
{
	struct cmdu_ackq *q = (struct cmdu_ackq *)cmdu_q;
	struct cmdu_ackq_entry *msg = NULL;

	msg = cmdu_ackq_lookup(cmdu_q, type, mid, dest);
	if (msg) {
		dbg("Duplicate: type = 0x%04x  mid = %hu origin = " MACFMT "\n",
		     type, mid, MAC2STR(dest));
		return -1;
	}

	msg = cmdu_ackq_create_msg(type, mid, dest, timeout, resend_cnt, cookie);
	if (msg) {
		int idx = cmdu_ackq_hash(type, mid, dest);

		pthread_mutex_lock(&q->qlock);
		hlist_add_head(&msg->hlist, &q->table[idx]);

		q->pending_cnt++;
		dbg("    ENQ:        type = 0x%04x  mid = %hu origin = " MACFMT "\n",
		     type, mid, MAC2STR(dest));

		if (timer_pending(&q->ageing_timer)) {
			loud("Timer pending ===\n");
			if (timercmp(&q->next_tmo, &msg->ageing_tmo, >)) {
				q->next_tmo.tv_sec = msg->ageing_tmo.tv_sec;
				q->next_tmo.tv_usec = msg->ageing_tmo.tv_usec;

				timer_set(&q->ageing_timer, msg->ageing_time);
				loud("Adjust ageout timer ===\n");
			}
		} else {
			loud("Start ageout timer ===\n");
			q->next_tmo.tv_sec = msg->ageing_tmo.tv_sec;
			q->next_tmo.tv_usec = msg->ageing_tmo.tv_usec;
			timer_set(&q->ageing_timer, msg->ageing_time);
		}

		pthread_mutex_unlock(&q->qlock);
		return 0;
	}

	return -1;
}

int cmdu_ackq_dequeue(void *cmdu_q, uint16_t type, uint16_t mid, uint8_t *src, void **cookie)
{
	struct cmdu_ackq *q = (struct cmdu_ackq *)cmdu_q;
	struct cmdu_ackq_entry *msg = NULL;
	int idx;

	msg = cmdu_ackq_lookup(cmdu_q, type, mid, src);
	if (!msg) {
		dbg("DROP! CMDU not found: type = 0x%04x  mid = %hu origin = " MACFMT "\n",
		     type, mid, MAC2STR(src));
		return -1;
	}

	idx = cmdu_ackq_hash(type, mid, src);

	pthread_mutex_lock(&q->qlock);
	hlist_del(&msg->hlist, &q->table[idx]);
	q->pending_cnt--;
	pthread_mutex_unlock(&q->qlock);

	dbg("DEQ: type = 0x%04x  mid = %hu origin = " MACFMT "\n",
	     type, mid, MAC2STR(src));


	/* After returning cookie back to user, we can safely delete the msg */
	if (cookie)
		*cookie = msg->cookie;
	msg->cookie = NULL;

	cmdu_ackq_delete_msg(q, msg);

	return 0;
}
